#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This program is a web based interface to the KPY part of 
# the reporting automation system for the monthly highlights.

import datetime
from calendar import monthrange
import bottle     # added for bottle.app()
from bottle import route, request, template, run
from bottle import static_file
from bottle import debug

# Required for the KPY part of this program.
import argparse
import os, sys, imaplib
import interactions.interactions as interactions
import configparser 

# TODO this needs to be set differently on the server.
# Probably /var/www/kpy/views
bottle.TEMPLATE_PATH.insert(0,'/home/mlake/git/KPY/views')

TEST = True
version = '2018.05.18'

now = datetime.datetime.now().strftime('%Y.%m.%d at %I:%M:%S %P')
months = ['January','February','March','April','May','June','July',\
    'August','September','October','November','December']

###########
# Functions
###########

def readconf():
    '''
    This reads the configuration file.
    '''

    conf = configparser.ConfigParser()
    conf.read('autoreport_web.conf')
    allowed = conf.get('users','allowed').split()

    return allowed


def check_login(email, password):
    '''
    TODO Checks a users login to this page. We dont want just anyone using this page.
    '''
    return True


def format_query(year, month, months):
    '''
    This takes the year and month as strings like '2017' and 'April' and 
    returns a formatted string like 'SINCE 01-Apr-2017 BEFORE 30-Apr-2017'.
    This stupid format is required by the imap library.
    Note that the args year and month come in as strings.
    '''

    # This returns the abbreviated month name from the full month name, 
    # e.g 'Jan' from 'January'.
    # %B is full month name, %b is abbreviated month name.
    abbr_month = datetime.datetime.strptime(month,'%B').strftime('%b')

    # To get the number of days in a month we use monthrange; 
    # e.g. monthrange(2018,1) which returns a tuple (0, 31).
    month_index = months.index(month) + 1
    first_day = 1 
    last_day = monthrange(int(year), month_index)[1]

    since = 'SINCE %d-%s-%s' % (first_day, abbr_month, year)
    before = 'BEFORE %d-%s-%s' % (last_day, abbr_month, year)
    query = '%s %s' % (since, before)

    return query


def run_engagement_individual(email, password, year, month):
    '''
    Takes a users email, their password, a year and month (note all these come in as strings)  
    and runs the interactions module to create a path to an image of the email interactions.  
    
    TODO change args.query from hard coded to getting from web app interface.
    "SINCE 01-Apr-2017 BEFORE 03-May-2017" to "2017.05.01" and "2017.05.07"
    (since, start_str, before, end_str) = self.args.query.split(' ')
    '''
    
    parser = argparse.ArgumentParser()
    args = parser.parse_args()

    # This is the users' username and password obtained from the Home page.
    args.username = email 
    args.password = password

    # The LDAP query also needs a password. This is not here but is set as a 
    # shell environment variable which interactions/interactions.py will read.

    # This handles users that do not have "Sent" as their sent malbox. 
    # If they do use "Sent" then do not list them here. 
    mailbox_mapping = {
        'Mike.Lake@uts.edu.au': 'Sent', 
        'Joe.Bloggs@uts.edu.ay': 'Out'
    }

    if args.username in mailbox_mapping.keys():
        args.folder = mailbox_mapping[args.username]
    else: 
        args.folder = 'Sent'

    args.from_unit = None
    args.ignore_headers = False
    args.server = 'outlook.office365.com'
    args.ldap = 'ldap.uts.edu.au'
    args.ldap_proxy = 'cn=eResearch Proxy Agent,ou=System Users,o=UTS'
    args.unit_string = 'utsUnitLevel1'
    args.unit_string2 = 'utsFaculty'
    args.mapping = 'groups.json'
    #args.query = 'SINCE 03-Apr-2017 BEFORE 03-May-2017'
    args.query = format_query(year, month, months)

    iaction = interactions.Interactions(args)
    iaction.deal_with_the_emails()
    image_path = iaction.report()
    iaction.close()

    return image_path


#########################
# Routes are defined here
#########################

@route('/static/<filename:path>')
def serve_static(filename):
    return static_file(filename, root='/home/mlake/git/KPY/static')
    #root_path = os.path.dirname(filename)
    #print ('ROOT=', root_path)
    #return static_file(filename, root=root_path)


@route('/')
def start():
    '''
    The main Home page.
    '''
    
    # Get this year and this month so the drop down boxes can be automatically 
    # selected for the current date.
    #thisyear  = datetime.datetime.now().strftime('%Y')
    #thismonth = datetime.datetime.now().strftime('%B')

    return template('start_page', now=now, months=months)


@route('/about', method='GET')
def about():
    ''' 
    An "about" page shows information about this software.
    '''
    versions = {}
    versions['this']   = version
    versions['bottle'] = bottle.__version__

    return template('about_page', versions=versions )


@route('/process', method='POST')
@route('/process/', method='POST')
def process():
    '''
    The process page is invoked by a POST query from the Start page.
    '''
    now = datetime.datetime.now().strftime('%Y.%m.%d at %I:%M:%S %P')
    email = request.forms.get('email')
    password = request.forms.get('password')
    month = request.forms.get('month')  # This is a string.
    year = request.forms.get('year')    # This is a string.

    if not email:
        return template('start_page', now=now, months=months)

    if not password:
        return template('start_page', now=now, months=months) 

    if not month:
        return template('start_page', now=now, months=months)

    # If the user is not in the list of allowed users then no-go.
    if email not in readconf():
        return template('start_page', now=now, months=months, error=email)

    if check_login(email, password):
        img_path = run_engagement_individual(email, password, year, month)
        return template('process_page', email=email, img_path=img_path, month=month)
    else:
        return template('start_page', now=now, months=months) 


@route('/reports', method='GET')
def reports():
    '''
    TODO A page to show reports.
    '''

    return template('reports_page')


# Run the application
if TEST:
    # If debug(True) then Exception and Traceback output
    # will go to the web browser screen.
    debug(True)
    # Use this to run using Bottle's inbuilt server.
    run(host='localhost', port=8080, reloader=True)
    #run(host='138.25.37.49', port=8080, reloader=True)
else:
    debug(False)
    # This site says to use bottle.app()
    # https://groups.google.com/forum/#!topic/bottlepy/wRfgm4obLXk
    application = bottle.app()

