"""
This module provides a class for interacting with the GitLab API.
"""

import requests, base64, yaml, dateutil

class Gitlab(object):
    """Object representing a connection to GitLab.

Attributes:
    url (str): base URL of the GitLab server
    api_key (str): secret API key
    metadata_files (str): filename where extra project metadata is sourced

"""
    def __init__(self, cf):
        """Create a GitLab object.

Args:
    url (str): base GitLab url
    api_key (str): secret API key
    metadata_files ([str]): list of filenames to look in for metadata
"""
        self.url = cf['url']
        self.api_key = cf['api_key']
        self.metadata_files = cf['metadata_file']
    
    def api(self, pid=None, path=None, args=None):
        """Low-level, internal method for an api call.

Don't call this directly, use one of the following methods instead:

    * projects() - get all project ids
    * project(id) - get basic data for a project
    * project_metadata(id) - get the extended metadata for a project
    * milestones(id) - get milestones and issues for a project
        
Args:
    pid (int): project id
    path (str): path
    args (dict): cgi parameters

Returns:
    The return value of the API call, decoded from JSON, on success.
    The data type depends on the nature of the API call - see the
    methods listed above for details.
    If the request fails, returns None.

"""
        url = self.url
        if pid:
            url += '/%d' % pid
        if path:
            url += '/' + path
        url += "?private_token=%s" % self.api_key
        if args:
            for key, value in args.items():
                url += "&%s=%s" % ( key, value )
        #print(url)
        response = requests.get(url)
        if response.status_code == 200:
            return response.json()
        else:
            return None

    def projects(self):
        """Fetch the list of projects from GitLab

Returns:
    (list) a list of dicts, each representing a project, and containing the
    decoded JSON of everything exposed by the GitLab API. The important
    fields are:
        id (int): project id
        name (str): name
        description (str): description
"""
        precs = self.api(args={"per_page": "100"})
        print("Got records for %d projects" % (len(precs)))
        return precs
        
    def project_metadata(self, pid):
        """Look for a metadata block in a README.md file and decode it.

For the purposes of reporting, we're including a yaml block at the start
of each project's README.md file. This method looks for a readme file and
tries to decode the metadata block.

Returns:
    (dict) on success, returns a dict containg the metadata values
    (None) if the request failed, there was no readme file, there was no
    metadata block or the metadata parse failed.

The Gitlab config includes a value 'metadata_file' which is a list of
filenames: each of these will be fetched and parsed in order. It's a list
so that the script can cope with variant capitalisations.
        """
        mdcontent = self.try_metadata(pid)

        if mdcontent:
            if 'content' in mdcontent:
                content = base64.b64decode(mdcontent['content'])
                chunks = content.decode('utf-8').split("---\n")
                if len(chunks) > 1:
                    try:
                        metadata = yaml.load(chunks[1])
                        return metadata
                    except yaml.YAMLError as exc:
                        if hasattr(exc, 'problem_mark'):
                            mark = exc.problem_mark
        return None
   
    def try_metadata(self, pid):
        """Tries the different metadata filenames until it gets a result.

Loops through the filenames in the config variable 'metadata_file' until
it finds one.

Returns:
    file content (dict): the content is the dict value for 'content', base64
    encoded
    None: if none of the files were found
"""
        
        for file in self.metadata_files:
            md = self.api(pid=pid, path='repository/files', args={ 'file_path': file, 'ref': 'master'} )
            if md:
                return md
        return None


    def date(self, dstr):
        """Splits a string date and converts it to a datetime date

Args:
    dstr (str): a string of the form 'YYYY-MM-DD'

Returns:
    date (datetime.date): a date
    None: if the string couldn't be split into three fields
"""
        bits = dstr.split('-')
        if len(bits) < 3:
            return None
        y, m, d = ( bits[0], bits[1], bits[2] )
        return datetime.date(y, m, d)

    def timestamp(self, tstr):
        return self.date(tstr[:10])

    def milestones(self, pid):
        """Returns a list of project milestones with issues"""
        milestones = []
        ms = self.api(pid=pid, path="milestones")
        for m in ms:
            if 'due_date' in m:
                if m['due_date']:
                    m['due_date'] = dateutil.parser.parse(m['due_date']).date()
            else:
                m['due_date'] = None
            issues = self.api(pid=pid, path="milestones/%s/issues" % m['id'])
            if issues:
                m['issues'] = issues
            else:
                m['issues'] = []
            milestones.append(m)
        return milestones
