# Objects for the ServiceConnect jobs spreadsheet

import xlrd
import pandas as pd
import matplotlib.pyplot as plt
import re, datetime, os.path, calendar, sys

EXCEL_DATEMODE = 0


class Jobs(object):
    def __init__(self, hr, cf):
        self.hr = hr
        self.cf = cf

    def excel_date(self, d):
        if not d:
            return None
        try:
            y, m, d, _, _, _ = xlrd.xldate_as_tuple(d, EXCEL_DATEMODE)
            return datetime.date(y, m, d)
        except ValueError:
            return None

    def state_case(self, s):
        return s[0].upper() + s[1:].lower()

    def in_this_month(self, d):
        "Return true if a date falls in the reporting month"
        if d:
            if d >= self.hr.period['start'] and d <= self.hr.period['end']:
                return True
        return False


    # FIXME - need to separate Closed jobs into two categories:
    #            - jobs which were new and closed this month
    #            - jobs which were already open and closed this month
    # This should make it easier to reconcile with the monthly history
    # chart.
    
    
    def load(self):
        # Open the service connect spreadsheet
        sc_xls_wb = xlrd.open_workbook(self.cf['report'])
        ws = sc_xls_wb.sheet_by_index(0)

        closed_re = re.compile(self.cf['closed_re'])
        self.jobs = {}        # jobs this month by faculty
        self.history = {}     # all jobs by month
        self.org_units = []
        cst = self.cf['state_col']
        cop = self.cf['opened_col']
        ccl = self.cf['closed_col']
        cou = self.cf['org_unit_col']
        cdesc = self.cf['desc_col']

        # set up filter for months to count
        
        nowy = self.hr.period['start'].year
        nowm = self.hr.period['start'].month
        theny = nowy
        thenm = nowm - self.cf['history'] + 1
        if thenm < 1:
            thenm += 12
            theny -= 1
        first = datetime.date(theny, thenm, 1)
        self.month_history = self.month_list(datetime.date(theny, thenm, 1), self.hr.period['start'])        

        #print("Type,Date,State,Desc")
                    
        for row in range(1, ws.nrows):
            state = self.state_case(ws.cell(row, cst).value)
            desc = ws.cell(row, cdesc).value
            dopen = self.excel_date(ws.cell(row, cop).value)
            dclosed = self.excel_date(ws.cell(row, ccl).value)
            #print("job-{},{},{},{}".format(row,dclosed,state,desc))
            if dopen >= self.hr.period['end']:
                #print("Date {} past {}".format(dopen, self.hr.period['end']))
                continue
            if closed_re.search(state):
                if dclosed:
                    if self.in_this_month(dclosed):
                        if self.in_this_month(dopen):
                            state = 'New, closed same month'
                        else:
                            state = 'Closed'
                    else:
                        state = 'Old'
                    #closed_when = "{y:04d}{m:02d}".format(m=dclosed.month, y=dclosed.year)
            # add jobs which are current, or closed this month, to
            # the chart by org unit
            if not state == 'Old':
                org_unit = self.hr.faculty(ws.cell(row, cou).value)
                self.add_job_org_unit(state, org_unit, desc)
            self.add_job_monthly(dopen, dclosed, state, desc)
                
        self.totals = { 'Open': 0, 'New': 0, 'Closed': 0 }

        for status, facs in self.jobs.items():
            count = sum(facs.values())
            if status in [ 'Closed', 'New' ]:
                self.totals[status] += count
            else:
                self.totals['Open'] += count


    def add_job_org_unit(self, state, org_unit, desc):
        """Add a job to the chart-by-org-units"""
        if org_unit not in self.org_units:
            self.org_units.append(org_unit)
        if state not in self.jobs:
            self.jobs[state] = {}
        if org_unit not in self.jobs[state]:
            self.jobs[state][org_unit] = 0
        self.jobs[state][org_unit] += 1
        #print("org,{},{},{}".format(org_unit, state, desc))

    def add_job_monthly(self, dopen, dclosed, state,desc):
        """Add a job to each of the months in which it's open"""
        if dclosed or state == 'Closed':
            months = self.month_list(dopen, dclosed)
            #print("monthly,{},{},{}".format('-'.join(months),state,desc))
            if not months:
                print("Bad months in {}".format(desc))
                return
            if len(months) == 1:
                self.add_job_month(desc, months[0], 'New, closed same month')
            else:
                for m in months[:-1]:
                    self.add_job_month(desc, m, 'Open')
                self.add_job_month(desc, months[-1], 'Closed')
        else:
            months = self.month_list(dopen, self.hr.period['end'])
            #print("monthly,{},{},{}".format('-'.join(months),state,desc))
            if not months:
                print("Bad months in {}".format(desc))
                return
            for m in months:
                self.add_job_month(desc, m, 'Open')

    def add_job_month(self, desc, month, state):
        """Add a job/state to a monthly total"""
        if month in self.month_history:
            #print("month,{},{},{}".format(month, state, desc))
            if state not in self.history:
                self.history[state] = {}
            if month not in self.history[state]:
                self.history[state][month] = 0
            self.history[state][month] += 1

    def yearmonth(self, m, y):
        """Standard format for using months as dict keys"""
        return "{y:04d}{m:02d}".format(m=m,y=y)
        
    def month_list(self, start, end):
        """Take two dates and return a list of YYYYMM strings spanning them"""
        
        if end < start:
            print("Bad date range, start={}, end={}".format(start, end))
            return None
            sys.exit(-1)
        last = self.yearmonth(end.month, end.year)
        year = start.year
        month = start.month
        done = False
        mlist = []
        while not done:
            current = self.yearmonth(month, year)
            if current == last:
                done = True
            if current == '202101':
                print("went too far!!!")
                return None
            mlist.append(current)
            month += 1
            if month == 13:
                month = 1
                year += 1
        return mlist

    # view methods for template: open/new/finished/closed
    def open(self):
        return self.totals['Open']

    def new(self):
        return self.totals['New']

    def finished(self):
        return self.totals['Finished']

    def closed(self):
        return self.totals['Closed']
    
                
    def main_chart(self): 
        # get the colour dict into a list 
        sc_colours = [ self.cf['colours'][s] for s in self.cf['order'] ]
        plt.figure(figsize=(40,10)) 
        scdf = pd.DataFrame(self.jobs, columns=self.cf['order'])
        self.dump_csv(scdf, 'main.csv')
        scdf.to_json(os.path.join(self.hr.cf['store'], 'service_connect.json'))
        scdf.plot(kind='bar', stacked=True, figsize=(12,9), color=sc_colours)
        chartfile = os.path.join(self.hr.cf['outdir'], self.cf['main_chart']) 
        plt.tight_layout()
        plt.savefig(chartfile)
        print("Wrote jobs chart to %s" % chartfile)

    def history_chart_old(self):
        # chart of jobs closed per month, starting at 'earliest'
        plt.figure(figsize=(40, 10))
        months = list(self.closed_history.keys())
        months.sort()
        months = months[-(self.cf['history'] + 1):-1]
        counts = [ self.closed_history[m] for m in months ]
        sc_colours = [ self.cf['colours'][s] for s in self.cf['order'] ]
        sc_colours.reverse()
        labels = [ calendar.month_abbr[int(m[4:6])] for m in months ]
        scdf = pd.Series(data=counts, index=labels)
        scdf.plot(kind='bar', figsize=(12, 9), color=sc_colours[0])
        chartfile = os.path.join(self.hr.cf['outdir'], self.cf['history_chart'])
#        plt.xticks(months, labels)
        plt.tight_layout()
        plt.savefig(chartfile)
        print("Wrote clearance history chart to %s" % chartfile)

    def new_history_chart(self):
        # chart of jobs closed per month, starting at 'earliest'
        plt.figure(figsize=(40, 10))
        months = self.month_history
        sc_colours = [ self.cf['history_colours'][s] for s in self.cf['history_order'] ]
        labels = [ calendar.month_abbr[int(m[4:6])] for m in months ]
        scdf = pd.DataFrame(data=self.history, columns=self.cf['history_order'])
        self.dump_csv(scdf, 'history.csv')
        ax = scdf.plot(kind='bar', stacked=True, figsize=(12, 9), color=sc_colours)
        ax.set_xticklabels(labels)
        chartfile = os.path.join(self.hr.cf['outdir'], self.cf['history_chart'])
        #plt.xticks(months, labels, rotation='vertical')
        plt.tight_layout()
        plt.savefig(chartfile)
        print("Wrote clearance history chart to %s" % chartfile)

        
    def dump_csv(self, df, file):
        path = os.path.join(self.hr.cf['outdir'], file)
        with open(path, 'w') as f:
            df.to_csv(f)
