"""
This module provides classes modelling projects, milestones and issues,
and a top-level Projects class which collects stats and makes them available
for the reports template.
"""


import dateutil
from kpy.gitlab import Gitlab
import pandas as pd
import matplotlib.pyplot as plt
import os.path

class Issue(object):
    """Object representing a project issue.

Attributes:
    id (str): issue id in GitLab
    name (str): issue name
    status (str): one of 'new', 'finished', 'open' or 'closed'.
    updated (date): date last updated

An issue is 'new' or 'closed' if it was opened or closed in the current
reporting period; otherwise it's 'open' or 'finished'.
    """

    def __init__(self, id, name, status, updated):
        """Create a project issue

Args:
    id (str): issue id
    name (str): issue name
    status (str): issue status
    updated (str): last updated (as a string)
    """
        self.id = id
        self.name = name
        self.status = status
        if updated: 
            self.updated = dateutil.parser.parse(updated).date()
        else:
            self.updated = None

            

class Milestone(object):
    """Object representing a project milestone.

Attributes:
    project (Project): project to which it belongs
    title (str): milestone title
    due_date (str): due date
    iopen (int): count of open issues
    inew (int): count of issues opened this month
    ifinished (int): count of closed issues
    iclosed (int): count of issues closed this month
    issues ([ Issue ]): list of issues
    active (bool): did any activity (open/close issues) happen this month?
"""

    
    def __init__(self, project, title, due_date, issues):
        """Create a new Issue

Args:
    project (Project): Project object to which this milestone belongs
    title (str): milestone title
    due_date (str): project du date
    issues ([ {} ]): list of dicts describing issues.

The list of issues should be in the format returned by the milestones
method from Gitlab.
"""
        self.project = project
        self.title = title
        self.due_date = due_date
        self.iopen = 0
        self.inew = 0
        self.ifinished = 0
        self.iclosed = 0
        self.issues = []
        self.active = False
        for i in issues:
            issue = Issue(i['id'], i['title'], i['state'], i.get('updated_at', None))
            issue.current = False
            if issue.updated:
                if issue.updated > self.project.period['start']:
                    issue.current = True
            if issue.current:
                if issue.status == 'opened':
                    self.inew += 1
                else:
                    self.iclosed += 1
            else:
                if issue.status == 'opened':
                    self.iopen += 1
                else:
                    self.ifinished += 1
            self.issues.append(issue)
        if self.inew + self.iclosed > 0:
            self.active = True


class Project(object):
    """Object representing a project.

Attributes:
    hr (HighlightsReport): the highlights report
    cf (dict): dictionary of config settings
    period (dict): start and end dates of reporting period
    pid (str): project id in GitLab
    name (str): project name
    description (str): project description
    ptype (str): project type
    metadata (dict): the metadata block at the top of the project's README.md
    report_status (str): project status one-liner for report
    status_updated (date): date that report_status was updated
    start_date (date): project start date
    close_date (date): project close date (if closed)
    hide (bool): flag whether to hide from report (based on type)
    traffic_light (str): project traffic-light status (red/amber/green) 
    new (bool): was the project created this month?
    closed (bool): was the project closed this month?
    iopen (int): count of open issues
    inew (int): count of issues opened this month
    ifinished (int): count of closed issues
    iclosed (int): count of issues closed this month
    overdue (int): count of overdue issues
    """
    def __init__(self, hr, cf, pdict, metadata):
        """Create a project based on values from GitLab

Args:
    hr (HighlightsReport): the highlights report
    cf (dict): the 'Projects' section of the config
    pdict (dict): the project dict from GitLab
    metadata (dict): the metadata dict from GitLab
"""
        self.cf = cf
        self.hr = hr
        self.period = hr.period
        self.pid = pdict['id']
        self.name = pdict['name']
        self.description = pdict['description']
        self.metadata = metadata
        self.ptype = self.metadata.get('type', '')
        if not self.ptype in self.cf['types']:
            self.ptype = None
            self.hide = True
            return
        self.start_date = None
        created = self.metadata.get('created_at')
        if created:
            self.start_date = gitlab.timestamp(self.created)
        self.new = False
        if self.start_date:
            if self.start_date >= self.period['start']:
                self.new = True
        self.closed = False
        self.close_date = self.metadata.get('actual-close-date')
        if self.close_date:
            try:
                if self.close_date >= self.period['start']:
                    self.closed = True
            except TypeError as t:
                self.closed = False
        self.traffic_light = self.metadata.get('traffic-light')
        self.status_updated = None
        if 'report-status-updated' in self.metadata:
            self.status_updated = self.metadata['report-status-updated']
            if 'report-status' in self.metadata and self.status_updated >= self.period['start']:
                self.report_status = self.metadata['report-status']
            else:
                self.report_status = None
            
    def date(self, dstr):
        """Convert a YYYY-MM-DD date string to a datetime.Date"""
        bits = dstr.split('-')
        if len(bits) < 3:
            return None
        y, m, d = ( bits[0], bits[1], bits[2] )
        return datetime.date(y, m, d)


    def percentage_done(self):
        """Calculate the project's percentage-done.

Percentage done = 100 * (closed/finised items) / (all items)
"""
        nissues = self.iopen + self.inew + self.iclosed + self.ifinished
        if nissues > 0:
            return int(100 * (self.iclosed + self.ifinished) / nissues)
        return 0
                    
    def add_milestones(self, mdicts):
        """Adds the milestone dicts from GitLab

Takes a list of milestone dictionaries from GitLab, uses them to build
a list of Milestone objects and calculates the stats for this project

Args:
    mdicts ([dict]) - list of dicts from GitLab
    
        """
        self.milestones = []
        for m in mdicts:
            milestone = Milestone(self, m['title'], m.get('due_date', None), m['issues'])
            self.milestones.append(milestone)
        self.iopen = 0
        self.inew = 0
        self.ifinished = 0
        self.iclosed = 0
        self.overdue = 0
        
        for m in self.milestones:
            self.iopen += m.iopen
            self.inew += m.inew
            self.ifinished += m.ifinished
            self.iclosed += m.iclosed
            if m.due_date:
                if m.due_date <= self.period['end']:
                    self.overdue += 1
    
    def active_milestones(self):
        """Returns a list of all milestones with activity this month"""
        return [ m for m in self.milestones if m.active ]
        

class Projects(object):
    """Object representing a collection of projects

A list of Project objects with methods to return totals for the report
template and render the projects chart.

Attributes:
    hr (HighlightsReport): the highlights report
    cf (dict): the 'Projects' section of the config
    issues (dict): issue totals: dict by status then projecttype
    projects ([Project]): list of Project objects

    """
    def __init__(self, hr, cf):
        """Create a Projects object.

Args:
    hr (HighlightsReport): the highlights report
    cf (dict): the config dict
"""
        self.hr = hr
        self.cf = cf
        self.issues = {
            'Open' : {},
            'Closed' : {},
            'New' : {}
            }
        self.projects = []
        for ptype in self.cf['types']:
            for status in self.issues.keys():
                self.issues[status][ptype] = 0
                
    def load(self):
        """Load the projects from GitLab and calculate totals."""
        gl = Gitlab(self.cf)
        pdicts = gl.projects()
        for pdict in pdicts:
            pmd = gl.project_metadata(pdict['id'])
            if pmd:
                project = Project(self.hr, self.cf, pdict, pmd)
                if project.ptype:
                    project.add_milestones(gl.milestones(pdict['id']))
                    self.add_project(project)
                    print("Added project '%s' (%d)" % (pdict['name'], pdict['id'] ))
                else:
                    print("Skipping project '%s' (%d)" % ( pdict['name'], pdict['id'] ))
                    print(project)
            else:
                print("Warning: couldn't read metadata for '%s' (%d)" % ( pdict['name'], pdict['id'] ))

                   
    def add_project(self, project):
        """Add a project object, incrementing the issue totals.

Args:
    project (Project): the project object
"""
        ptype = project.ptype
        self.issues['Open'][ptype] += project.iopen
        self.issues['New'][ptype] += project.inew
        self.issues['Closed'][ptype] += project.ifinished
        self.projects.append(project)

    def totals_by_type(self, ptype):
        """Return a dict of issues by status for a project type."""
        return { status: self.issues[status][ptype] for status in self.issues.keys() }
            
    def engagement(self):
        """Dict of Open/Closed/New issue totals for Engagement projects"""
        return self.totals_by_type('Engagement')

    def infrastructure(self):
        """Dict of Open/Closed/New issue totals for Infrastructure projects"""
        return self.totals_by_type('Infrastructure')

    def development(self):
        """Dict of Open/Closed/New issue totals for Development projects"""
        return self.totals_by_type('Development')
    
    def administration(self):
        """Dict of New/Closed/Open issue totals for Adminstration projects"""
        return self.totals_by_type('Administration')

    def totals(self):
        """Dict of New/Closed/Open issue totals across all types"""
        return { status: sum(self.issues[status].values()) for status in self.issues.keys() }

    def chart(self):
        """Writes the projects chart to the 'imgfile' in the config"""
        cf = self.cf
        cfg = self.hr.cf
        gl_colours = [ cf['colours'][s] for s in cf['order'] ]

        projdf = pd.DataFrame(self.issues, columns=cf['order'])
        projdf.to_json(os.path.join(cfg['store'], 'gitlab_projects.json'))
        projdf.plot(kind='bar', stacked=True, figsize=(12,9), color=gl_colours)
        outfile = os.path.join(cfg['outdir'], cf['imgfile'])
        plt.tight_layout()
        plt.savefig(outfile)
        print("Wrote projects chart to %s" % outfile)
